import re
from logging import getLogger

from uuidgen import uuidgen

LOG = getLogger('TALEnginePieces')

# We compile the TALES stuff we find using this.
import Expressions


EMPTY_HTML_TAGS = [
	# List of HTML tags with an empty content model; these are
	# rendered in minimized form, e.g. <img />.
	# From http://www.w3.org/TR/xhtml1/#dtds
	"base", "meta", "link", "hr", "br", "param", "img", "area",
	"input", "col", "basefont", "isindex", "frame",
	]

BOOLEAN_HTML_ATTRS = [
	# List of Boolean attributes in HTML that may be given in
	# minimized form (e.g. <img ismap> rather than <img ismap="">)
	# From http://www.w3.org/TR/xhtml1/#guidelines (C.10)
	"compact", "nowrap", "ismap", "declare", "noshade", "checked",
	"disabled", "readonly", "multiple", "selected", "noresize",
	"defer",
	]


attrfind = re.compile(
	r'\s*([a-zA-Z_][-.:a-zA-Z_0-9]*)(\s*=\s*'
	r'(\'[^\']*\'|"[^"]*"|[-a-zA-Z0-9./:;+*%?!&$\(\)_#=~]*))?')

NAME_RE = "[a-zA-Z_][a-zA-Z0-9_]*"

_attr_re = re.compile(r"\s*([^\s]+)\s+([^\s].*)\Z", re.S)
_subst_re = re.compile(r"\s*(?:(text|structure)\s+)?(.*)\Z", re.S)

def parseAttributeReplacements(arg):
	dict = {}
	for part in splitParts(arg):
		m = _attr_re.match(part)
		if not m:
			raise TALError("Bad syntax in attributes:" + `part`)
		name, expr = m.group(1, 2)
		if dict.has_key(name):
			raise TALError("Duplicate attribute name in attributes:" + `part`)
		dict[name] = expr
	return dict


def parseSubstitution(arg, position=(None, None)):
	m = _subst_re.match(arg)
	if not m:
		raise TALError("Bad syntax in substitution text: " + `arg`, position)
	key, expr = m.group(1, 2)
	if not key:
		key = "text"
	return key, expr

def splitParts(arg):
	# Break in pieces at undoubled semicolons and
	# change double semicolons to singles:
	args = arg.replace(";;","\0")
	parts = args.split(';')
	parts = map(lambda s: s.replace("\0", ";"), parts)
	if len(parts) > 1 and not parts[-1].strip():
		del parts[-1] # It ended in a semicolon
	return parts


class ExpressionLookupTable(object):
	__slots__ = ('globals', 'conditions', 'loops', 'contents',
				 'replaces', 'structures', 'defines', 'attributes',
				 'expr', 'onErrors', 'allExpr', 'slots', 'macros',
				 'slotDefines', 'useMacros', 'defineOrder')
	
	def __init__(self):
		self.globals = {}
		self.conditions = {}
		self.loops = {}
		self.contents = {}
		self.replaces = {}
		self.structures = []
		self.defines = {}
		self.defineOrder = []
		self.attributes = {}
		self.expr = {}
		self.onErrors = {}
		self.allExpr = {}
		self.slots = {}
		self.macros = {}
		self.slotDefines = {}
		self.useMacros = {}

	def addStructure(self, id):
		self.structures.append(id)

	def isStructure(self, id):
		return id in self.structures
		
	def addSlot(self, id, expr):
		self.slots[id] = expr
		self.allExpr[id] = expr
		
	def addSlotDefine(self, id, expr):
		self.slotDefines[id] = expr
		self.allExpr[id] = expr
 		
	def addMacro(self, id, block):
		self.macros[id] = block

	def addUseMacro(self, id, expr):
		self.useMacros[id] = expr
		self.allExpr[id] = expr

	def addGlobal(self, id, expr):
		self.defines[id] = expr
		self.allExpr[id] = expr
		self.globals[id] = expr
		self.defineOrder.append(id)
		
	def isGlobal(self, id):
		return self.globals.has_key(id)

	def addCondition(self, id, expr):
		self.conditions[id] = expr
		self.allExpr[id] = expr
		
	def addLoop(self, id, expr):
		self.loops[id] = expr
		self.allExpr[id] = expr

	def addExpr(self, id, expr):
		self.allExpr[id] = expr
		
	def addContent(self, id, expr):
		self.contents[id] = expr
		self.allExpr[id] = expr
		
	def addReplace(self, id, expr):
		self.replaces[id] = expr
		self.allExpr[id] = expr
		
	def addDefine(self, id, expr):
		self.defines[id] = expr
		self.allExpr[id] = expr
		self.defineOrder.append(id)

	def addAttribute(self, id, expr):
		self.attributes[id]=expr
		self.allExpr[id]=expr

	def addOnError(self, id, expr):
		self.onErrors[id]=expr
		self.allExpr[id]=expr

	def get(self, id):
		return self.allExpr[id]

	def getMacro(self, id):
		return self.macros[id]

# Common Page Element code, with default methods.
#
# Pages consist of Tags and Chunks (of text).
# A Tag can contain other Tags and Chunks.
#
class PageElem(object):
	__slots__ = ('subBlocks', 'isTag', 'isContainer', 'asContainer')

	# Reduce the number of blocks by converting suitable Tags into
	# Chunks.
	# Then by combining consecutive Chunks into a single Chunk
	# We do this to reduce the total amount of code we generate, and
	# therefore the total amount of code we need to execute, on the theory that
	# one call that writes a big chunk is cheaper than lots of little calls.
	#
	# Pages and Tags have subBlocks. Chunks do not.
	#
	def collapseSelf(self):
		if not self.subBlocks:
			return
		count = len(self.subBlocks)

		if count == 1:
			return
		
		for block in self.subBlocks:
			block.collapseSelf()

		newBlocks = []
		allChunks = True
		for block in self.subBlocks:
			if block.canBeText() and block.isTag:
				buffer = block.getBuffer()
				newBlock=Chunk(buffer, block.startLine, block.endLine)
				newBlocks.append(newBlock)
			elif block.isSlot:
				pass
			else:
				allChunks = False
				newBlocks.append(block)

		self.subBlocks = newBlocks

		if allChunks:
			buffers = []
			for b in self.subBlocks:
				buffers.append(b.getBuffer())

			buffer = ''.join(buffers)
			newBlock = Chunk(buffer,
							 self.subBlocks[0].startLine,
							 self.subBlocks[-1].endLine)
			self.subBlocks = [newBlock,]
		
	# Mainly used as a debugging aid to see the amount of reduction.
	def __len__(self):
		count = sum(map(len, self.subBlocks)) + 1 # add self..
		return count

	# Add a sub block
	def addBlock(self, block):
		self.subBlocks.append(block)

	# Can we be represented as a text block
	def canBeText(self):
		return True

	def checkAttributes(self, rawdata):
		endpos = len(rawdata)
		k = 0
		attrs = []
		while k < endpos:
			m = attrfind.match(rawdata, k)
			if not m:
				break
			attrname, rest, attrvalue = m.group(1, 2, 3)
			if not rest:
				attrvalue = None
			elif attrvalue[:1] == '\'' == attrvalue[-1:] or \
					 attrvalue[:1] == '"' == attrvalue[-1:]:
				attrvalue = attrvalue[1:-1]
				attrvalue = self.unescape(attrvalue)

##			if not ':' in attrname:
##				attrvalue = attrvalue.replace('%', '%%')
				
			attrs.append((attrname.lower(), attrvalue))
			k = m.end()

		self.attributes = attrs

		for a in self.attributes:
			b = a[0].split(':')
			if len(b) == 1:
				continue
			if b[0] in ('tal', 'metal'):
				self.hasTal = 1
				if b[1].lower() == 'content':
					self.asContainer = 1
				break

	def unescape(self, s):
		if '&' not in s:
			return s
		s = s.replace("&lt;", "<")
		s = s.replace("&gt;", ">")
		s = s.replace("&apos;", "'")
		s = s.replace("&quot;", '"')
		s = s.replace("&amp;", "&") # Must be last
		return s

	# Generate the python code for this page.
	def spewCode(self, pdepth = 0):
		depth = pdepth
		buffer = ''
		indent = '\t'*depth
		bufferList = []
		
		# We're not a tag, just return our buffer.
		if not self.isTag:
			indent = '\t'*depth
			buffer = buffer + indent + 'pt_write(\'%s\')\n'%(py_quote(self.buffer))
			return buffer

		if ( self.lookupTable.loops or
			 self.lookupTable.defines # or self.lookupTable.conditions
			 ):
			indent = '\t'*depth
			bufferList.append(indent + 'context.beginScope()\n')


		# tal:on-error -> try: except: block
		if self.lookupTable.onErrors:
			indent = '\t'*depth
			bufferList.append(indent + 'try:\n')
			depth = depth + 1

		# tal:repeat=
		if self.lookupTable.loops:
			indent = '\t'*depth			
			name, value = self.lookupTable.loops.items()[0]
			bufferList.append(indent + '%sIter=context.setRepeat("%s", getLoopExpr("%s"))\n'%(name, name, value))
			bufferList.append(indent + 'while %sIter.next():\n'%(name))
			depth = depth + 1

		# tal:define=
		if self.lookupTable.defines:
			indent = '\t'*depth
			for name in self.lookupTable.defineOrder:
				value = self.lookupTable.defines[name]
				if self.lookupTable.isGlobal(name):
					bufferList.append(indent + 'context.setGlobal("%s", getRawExpr("%s", context))\n'%(name, value))
				else:
					bufferList.append(indent + 'context.setLocal("%s", getRawExpr("%s", context))\n'%(name, value))

		# tal:condition=
		if self.lookupTable.conditions:
			indent = '\t'*depth
			name, value = self.lookupTable.conditions.items()[0]
			bufferList.append(indent + 'if getBooleanExpr("%s", context):\n'%(name))
			depth = depth + 1
			indent = '\t'*depth
			

		# we're not tal:omit-tag or tal:replace e.g.
		if self.printable:
			indent = '\t'*depth
			arglist = []
			for a in self.attributes:
				if a[0] not in BOOLEAN_HTML_ATTRS:
					arglist.append(py_quote('%s="%s"'%(a[0],a[1])))
				else:
					val = a[1]
					if val:
						try:
							val = int(a[1])
						except:
							pass
					if val:
						arglist.append( py_quote('%s'%(a[0])))

			if arglist:
				arglist = ' '+' '.join(arglist)+' '
			else:
				arglist = ''

			eArgs=''
			eArgList=[]
			if self.lookupTable.attributes:
				for name, value in self.lookupTable.attributes.items():
					if not name in BOOLEAN_HTML_ATTRS:
						eArgs += '%s'
						eArgList.append("getAExpr('%s', '%s', context)"%(value, name))
					else:
						eArgs += '%s'
						eArgList.append("getAExpr('%s', '%s', context, True)"%(value, name))
						
			if eArgList:
				# eArgList = '%('+','.join(eArgList)+')'
				eArgList = "'+" + '+'.join(eArgList) + "+'"
			else:
				eArgList = ''
												 
			if self.isContainer or self.asContainer:
				bufferList.append(indent + "pt_write('<%s%s%s>')\n"%(self.tagName, arglist,eArgList))
			else:
				bufferList.append(indent + "pt_write('<%s%s%s />')\n"%(self.tagName, arglist,eArgList))

		# metal:use-macro
		if self.lookupTable.useMacros:
			name, value = self.lookupTable.useMacros.items()[0]
			bufferList.append(indent + 'pt_write(execMacro("%s", context))\n'%(name))

		# metal:define-slot
		if self.lookupTable.slotDefines:
			indent = '\t'*depth
			name, value = self.lookupTable.slotDefines.items()[0]
			bufferList.append(indent + 'try:\n')
			bufferList.append(indent + '\tpt_write(execSlot("%s", context))\n'%(name))
			bufferList.append(indent + 'except NoSlotError:\n')
			depth = depth + 1


		# tal:contents
		if self.lookupTable.contents:
			indent = '\t'*depth
			name, value = self.lookupTable.contents.items()[0]
			bufferList.append(indent + 'pt_write(getCExpr("%s", context, \'%s\'))\n'%(value, py_quote(self.buffer)))

		else:
			for block in self.subBlocks:
				bufferList.append(block.spewCode(depth))

		# metal:define-slot
		if self.lookupTable.slotDefines:
			indent = '\t'*depth
			# Insert NOOP incase there's no obvious slot..
			bufferList.append(indent + 'pass\n')
			depth = depth - 1

		# do closing tag if we're a container type
		if (self.isContainer or self.asContainer) and self.printable:
			# bufferList.append(indent + "pt_write('</%s>')\n"%(self.tagName))
			# Try to minimise writes..
			rindex = bufferList[-1].rfind("')\n")
			if rindex >= 0:
				bufferList[-1] = bufferList[-1][:rindex] + "</%s>')\n"%(self.tagName)
			else:
				val = indent + "pt_write('</%s>')\n"%(self.tagName)
				bufferList.append(val)
			
		# close off tal:repeat
		if self.lookupTable.loops:
			depth = depth - 1
			indent = '\t'*depth

		# close off tal:condition
		if self.lookupTable.conditions:
			indent = '\t'*depth
			depth = depth - 1

		# close off tal:on-error
		if self.lookupTable.onErrors:			
			depth = depth - 1
			indent = '\t'*depth
			bufferList.append(indent + 'except:\n')
			depth = depth + 1
			indent = '\t'*depth
			name, value = self.lookupTable.onErrors.items()[0]
			bufferList.append(indent + 'pt_write(getCExpr("%s", context, \'%s\'))\n'%(value, py_quote(self.buffer)))
			depth = depth - 1

		if ( self.lookupTable.loops or
			 self.lookupTable.defines # or self.lookupTable.conditions
			 ):
			indent = '\t'*depth
			bufferList.append(indent + 'context.endScope()\n')
			

		return ''.join(bufferList)

	def showBlockStructure(self, depth = 0):
		indent = '  '*depth
		LOG.info(indent + self.tagName)
		for block in self.subBlocks:
			block.showBlockStructure(depth + 1)

		if self.isContainer:
			LOG.info(indent + '/' + self.tagName)
			
class Chunk(PageElem):
	__slots__ = ('isTag', 'isMacro', 'isSlot', 'buffer', 'hasTal',
				 'asContainer', 'isContainer', 'usesMacro', 'tagName',
				 'startLine', 'endLine', 'parentBlock')
	def __init__(self, buffer, startLine=0, endLine=0):
		self.isTag = False
		self.isMacro = False
		self.isSlot = False
		self.buffer = buffer
		self.hasTal = False
		self.isContainer = False
		self.asContainer = False
		self.usesMacro = False		
		self.tagName="[Block of text]"
		self.startLine=startLine
		self.endLine=endLine

	# You can't make a chunk any smaller
	def collapseSelf(self):
		pass

	# I am my own buffer.
	def getBuffer(self, depth = 0):
		return self.buffer

	def addBuffer(self, buffer):
		self.buffer += buffer
			
class Tag(PageElem):
	__slots__ = ('isTag', 'isMacro', 'isSlot', 'buffer', 'hasTal',
				 'asContainer', 'isContainer', 'usesMacro', 'subBlocks',
				 'tagName', 'startLine', 'endLine', 'lookupTable',
				 'isSlotDefine', 'printable', 'attributes', 'parentBlock')
	
	def __init__(self, buffer, startLine=0, endLine=0):
		self.isMacro = False
		self.isSlot = False
		self.isSlotDefine = False
		self.hasTal = False
		self.isTag = True
		self.isContainer = False
		self.asContainer = False
		self.printable = True
		self.usesMacro = False
		self.subBlocks=[]
		self.startLine = startLine
		self.endLine = endLine
		self.lookupTable=ExpressionLookupTable()

		tagAndAttributes = buffer[1:-1].strip().split()
		
		self.tagName = tagAndAttributes[0].lower()

		# Sometimes we get spares when there's 0 content in a container
		slashIndex = self.tagName.find('/')
		if slashIndex > 0:
			self.tagName=self.tagName[slashIndex:]

		self.attributes = []

		if len(tagAndAttributes)>1:
			attributes=' '+' '.join(tagAndAttributes[1:])+' '
			self.checkAttributes(attributes)

		# Stupid metal and tal namespace blocks... hate hate hate
		if self.tagName[:5]=='metal':
			self.isContainer = True
			self.isTag = True
			self.printable = False
			newAttributes=[]
			for k,v in self.attributes:
				if k[:5] != 'metal':
					newAttributes.append(('metal:'+k,v))
				else:
					newAttributes.append((k,v))
			del self.attributes
			self.attributes=newAttributes
			self.hasTal = True
		elif self.tagName[:3]=='tal':
			self.isContainer = True
			self.isTag = True
			self.printable = False
			newAttributes=[]
			for k,v in self.attributes:
				if k[:3] != 'tal':
					newAttributes.append(('tal:'+k,v))
				else:
					newAttributes.append((k,v))
					
			del self.attributes
			self.hasTal = True
			self.attributes=newAttributes

		if self.tagName not in EMPTY_HTML_TAGS:
			self.isContainer = True

		if buffer[-2:]=='/>':
			self.isContainer = False

	def getSubBuffer(self, depth = 0):
		buffers = []
		for block in self.subBlocks:
			buffers.append(block.getBuffer(depth+1))

		return ''.join(buffers)
		
	# My buffer is me and my sub buffers.
	def getBuffer(self, depth = 0):
		arglist = ''
		for a in self.attributes:
			arglist = arglist + ' %s="%s"'%(a[0],a[1])

		#indent = '  '*depth'
		indent = ''
		buffer = indent		
		if self.isContainer or self.asContainer:
			buffer = buffer + "<%s%s>"%(self.tagName, arglist)
		else:
			buffer = buffer + "<%s%s />"%(self.tagName, arglist)


		buffer += self.getSubBuffer(depth)

		if self.isContainer or self.asContainer:
			buffer = buffer + indent + "</%s>"%(self.tagName)
		return buffer

	# I can be a text block iff
	# I have no tal
	# AND I have no subBlocks
	# OR  I have subblocks, and they can all be text blocks too
	#
	def canBeText(self):
		if self.hasTal:
			return False

		if not self.subBlocks:
			return True

		for block in self.subBlocks:
			if not block.canBeText():
				return False
		return True

	def doAttributes(self, lookupTable, name, value):
		newAttrs = parseAttributeReplacements(value)
		for name, expr in newAttrs.items():
			id = getID()
			cexpr = Expressions.getEngine().compile(expr)
			lookupTable.addAttribute(id, cexpr)
			self.lookupTable.addAttribute(name, id)
			self.cleanAttributes(name)

	def cleanAttributes(self, name):
		self.attributes=filter(lambda a, name=name:a[0]!=name, self.attributes)
		
	def doRepeat(self, lookupTable, name, value):
		m = re.match("(?s)\s*(%s)\s+(.*)\Z" % NAME_RE, value)
		if not m:
			raise TALError, ("invalid repeat syntax: " + `value`, 0)
		name, expr = m.group(1, 2)
		cexpr = Expressions.getEngine().compile(expr)
		id = getID()
		lookupTable.addLoop(id, cexpr)
		self.lookupTable.addLoop(name, id)
		self.lookupTable.addExpr(id, cexpr)

	def doCondition(self, lookupTable, name, value):
		try:
			cexpr = Expressions.getEngine().compile(value)
			id = getID()
			lookupTable.addCondition(id, cexpr)
			self.lookupTable.addCondition(id, cexpr)
		except:
			print "Could not compile expression: Lines %d-%d"%(self.startLine, self.endLine)
			raise
		
	def doContent(self, lookupTable, name, value):
		key, expr = parseSubstitution(value)		
		cexpr = Expressions.getEngine().compile(expr)
		id = getID()
		lookupTable.addContent(id, cexpr)
		self.lookupTable.addContent(id, id)
		self.lookupTable.addExpr(id, cexpr)
		if key == 'structure':
			self.lookupTable.addStructure(id)
			lookupTable.addStructure(id)
		
	def doDefine(self, lookupTable, name, value):
		for part in splitParts(value):
			m = re.match(
				r"(?s)\s*(?:(global|local)\s+)?(%s)\s+(.*)\Z" % NAME_RE, part)
			if not m:
				raise TALError("invalid define syntax: " + `part`,
							   self.position)
			scope, name, expr = m.group(1, 2, 3)
			scope = scope or "local"
			cexpr = Expressions.getEngine().compile(expr)
			id = getID()
			if scope == 'local':
				lookupTable.addDefine(id, cexpr)
				self.lookupTable.addDefine(name, id)
			else:
				lookupTable.addGlobal(id, cexpr)
				self.lookupTable.addGlobal(name, id)
				
	def doOnError(self, lookupTable, name, value):
		cexpr = Expressions.getEngine().compile(value)
		id = getID()
		lookupTable.addOnError(id, cexpr)
		self.lookupTable.addOnError(id, id)
		self.lookupTable.addExpr(id, cexpr)
		
	def doDefineSlot(self, lookupTable, name, value):
		lookupTable.addSlotDefine(value, value)
		self.lookupTable.addSlotDefine(value, value)
		self.isSlotDefine = 1
		self.printable = 0

	def doDefineMacro(self, lookupTable, name, value):
		lookupTable.addMacro(value, self)
		self.lookupTable.addMacro(value, self)
		self.isMacro = 1

	def doUseMacro(self, lookupTable, name, value):
		cexpr = Expressions.getEngine().compile(value)
		id = getID()
		lookupTable.addUseMacro(id, cexpr)
		self.lookupTable.addUseMacro(id, cexpr)		
		self.usesMacro = 1
		self.printable = 0
		if self.subBlocks:
			self.subBlocks=[]

	def doFillSlot(self, lookupTable, name, value):
		lookupTable.addSlot(value, self)
		self.lookupTable.addSlot(value, self)
		self.isSlot = 1

	# Remove attributes filters for post-processing.
	def noCondition(self, item):
		return not item[0]=='tal:condition'

	def noContent(self, item):
		return not item[0]=='tal:content'

	def noReplace(self, item):
		return not item[0]=='tal:replace'

	def noOmitTag(self, item):
		return not item[0]=='tal:omit-tag'

	def noDefine(self, item):
		return not item[0]=='tal:define'

	def noRepeat(self, item):
		return not item[0]=='tal:repeat'

	def noAttributes(self, item):
		return not item[0]=='tal:attributes'

	def noDefineSlot(self, item):
		return not item[0]=='metal:define-slot'

	def noDefineMacro(self, item):
		return not item[0]=='metal:define-macro'

	def noUseMacro(self, item):
		return not item[0]=='metal:use-macro'

	def noFillSlot(self, item):
		return not item[0]=='metal:fill-slot'

	def noOnError(self, item):
		return not item[0]=='tal:on-error'

	# do things for metal-tags.
	def expandMetal(self, lookupTable, command, name, value):
		if command == 'define-macro':
			self.doDefineMacro(lookupTable, name, value)
			self.attributes=filter(self.noDefineMacro, self.attributes)
			self.printable=1
			
		if command == 'use-macro':
			self.doUseMacro(lookupTable, name, value)
			self.attributes=filter(self.noUseMacro, self.attributes)
			self.printable=0			
			self.subBlocks=[]
			
		if command == 'define-slot':
			self.doDefineSlot(lookupTable, name, value)
			self.attributes=filter(self.noDefineSlot, self.attributes)

			
		if command == 'fill-slot':
			self.doFillSlot(lookupTable, name, value)
			self.attributes=filter(self.noFillSlot, self.attributes)
			self.printable=0
			

	# do things for tal/metal tags.
	def expandTAL(self, lookupTable):

		
		# Expand our children first
		for block in self.subBlocks:
			if block.isTag:
				block.expandTAL(lookupTable)

		# If we don't have TAL we don't need to continue
		if not self.hasTal:
			return

		for attr in self.attributes:
			if len(attr) == 1:
				continue

			name, value = attr
			talType = ''
			command = ''
			try:
				talType,command = name.split(':')
			except:
				continue

			if talType == 'metal':
				self.expandMetal(lookupTable, command, name, value)
				continue
			
			if talType != 'tal':
				continue

			if command == 'attributes':
				self.doAttributes(lookupTable, name, value)
				self.attributes=filter(self.noAttributes, self.attributes)

			if command == 'define':
				self.doDefine(lookupTable, name, value)
				self.attributes=filter(self.noDefine, self.attributes)
				
			if command == 'condition':
				self.doCondition(lookupTable, name, value)
				self.attributes=filter(self.noCondition, self.attributes)

			if command == 'repeat':
				self.doRepeat(lookupTable, name, value)
				self.attributes=filter(self.noRepeat, self.attributes)
				
			if command == 'omit-tag':
				self.printable = 0
				self.attributes = filter(self.noOmitTag, self.attributes)

			if command == 'on-error':
				self.doOnError(lookupTable, name, value)
				self.attributes = filter(self.noOnError, self.attributes)

			if command == 'content' or command == 'replace':
				if value.find('default') >= 0:
					self.buffer = self.getSubBuffer()
				else:
					self.buffer = ''
				self.subBlocks = []
				self.doContent(lookupTable, name, value)
				if command == 'replace':
					self.printable = False
					self.attributes=filter(self.noReplace, self.attributes)
				else:
					self.attributes=filter(self.noContent, self.attributes)

# We are really just a simple case of Tag a mythical <page></page> container.
# There is only one page per PageTemplate.
class Page(PageElem):
	def __init__(self):
		self.isTag = False
		self.isMacro = False
		self.isSlot = False
		self.hasTal = False
		self.isContainer = False
		self.asContainer = False
		self.usesMacro = False		
		self.subBlocks=[]
		self.tagName="[Page Definition]"
		
	# Go through and build a lookup table for TAL expressions.
	def expandExpressions(self, lookupTable):
		for block in self.subBlocks:
			if block.isTag:
				block.expandTAL(lookupTable)

	# Throw up a buffer full of python code.
	def spewCode(self):
		buffers = []
		for block in self.subBlocks:
			buffers.append(block.spewCode())

		return ''.join(buffers)

#
# Used to generate unique ids for every expression we encounter, so we
# can make a lookup table.
#
def getID():
	return uuidgen()

# Make things safe to print inside python.
def py_quote(str):
	return str.replace("'", r"\'").replace('\n', r'\n')
